function engine = jtree_ndx_inf_engine(bnet, varargin)
% JTREE_NDX_INF_ENGINE Junction tree inference engine which precomputes indexes
% engine = jtree_ndx_inf_engine(bnet, ...)
%
% This is a version of jtree_inf_engine that sacrifices generality for speed.
% In particular,
% - we require all hidden nodes to be discrete
% - we require the nodes which are observed to always be the same;
% - we use 1D vectors instead of dpot objects
% - we pre-compute various indices to help speedup the manipulation of tables
% - we only allow the computation of marginals on families and singletons
%
% Additional optional parameters
% 'ndx_type ' - 'SD', 'D', 'B' ['SD']

% set default params
N = length(bnet.dag);
clusters = {};
root = N;
stages = { 1:N };
ndx_type = 'SD';

if nargin >= 3
  args = varargin;
  nargs = length(args);
  for i=1:2:nargs
    switch args{i},
     case 'clusters', clusters = args{i+1}; 
     case 'root',     root = args{i+1}; 
     case 'stages',   stages = args{i+1}; 
     case 'ndx_type',  ndx_type = args{i+1}; 
     otherwise,  
      error(['invalid argument name ' args{i}]);       
    end
  end
end

engine = init_fields;
engine = class(engine, 'jtree_ndx_inf_engine', inf_engine(bnet));

engine.ndx_type = ndx_type;

for n=1:N
  engine.bnet_parents{n} = parents(bnet.dag, n);
end

onodes = bnet.observed;
[engine.jtree, dummy, engine.cliques, B, w] = dag_to_jtree(bnet, onodes, stages, clusters);
engine.cliques_bitv = B;
engine.clique_weight = w;
C = length(engine.cliques);
engine.clpot = cell(1,C);

% Compute the separators between connected cliques.
[is,js] = find(engine.jtree > 0);
engine.separator = cell(C,C);
for k=1:length(is)
  i = is(k); j = js(k);
  engine.separator{i,j} = find(B(i,:) & B(j,:)); % intersect(cliques{i}, cliques{j});
end

% A node can be a member of many cliques, but is assigned to exactly one, to avoid
% double-counting its CPD. We assign node i to clique c if c is the "lightest" clique that
% contains i's family, so it can accomodate its CPD.

engine.clq_ass_to_node = zeros(1, N);
for i=1:N
  %c = clq_containing_nodes(engine, family(bnet.dag, i));
  clqs_containing_family = find(all(B(:,family(bnet.dag, i)), 2)); % all selected columns must be 1
  c = clqs_containing_family(argmin(w(clqs_containing_family)));  
  engine.clq_ass_to_node(i) = c; 
end

% Make the jtree rooted, so there is a fixed message passing order.
engine.root_clq = clq_containing_nodes(engine, root);
if engine.root_clq <= 0
  error(['no clique contains ' num2str(root)]);
end

[engine.jtree, engine.preorder, engine.postorder] = mk_rooted_tree(engine.jtree, engine.root_clq);

% collect 
engine.postorder_parents = cell(1,length(engine.postorder));
for n=engine.postorder(:)'
  engine.postorder_parents{n} = parents(engine.jtree, n);
end
% distribute
engine.preorder_children = cell(1,length(engine.preorder));
for n=engine.preorder(:)'
  engine.preorder_children{n} = children(engine.jtree, n);
end

if isempty(onodes)
  error('can only use ndx if you pre-specify the observed nodes');
  % otherwise we don't know what sizes to make the indices
end

pot_type = determine_pot_type(bnet, onodes);
if pot_type ~= 'd'
  error('jtree_ndx only works for discrete potentials')
end

ns = bnet.node_sizes;
engine.actual_node_sizes = ns;
ns(onodes) = 1;
engine.eff_node_sizes = ns;

% The indexing convention for index id's is as  follwos:
% let i = intersection(c1, c2) for cliques c1,c2
% mult_cl_by_sep_ndx_id(c1, c2):  pot(c1) *= pot(i)
% marg_cl_onto_sep_ndx_id(c1, c2): sum pot(c1) onto i

% collect
for n=engine.postorder 
  for p=engine.postorder_parents{n}
    %  seppot{p,n} = marginalize_pot(clpot{n}, engine.separator{p,n}, engine.maximize);
    engine.marg_cl_onto_sep_ndx_id(n,p) = add_ndx(engine.cliques{n}, engine.separator{p,n}, ns, ndx_type);
    %clpot{p} = multiply_by_pot(clpot{p}, seppot{p,n});
    engine.mult_cl_by_sep_ndx_id(p,n) = add_ndx(engine.cliques{p}, engine.separator{p,n}, ns, ndx_type);
  end
end

% distribute
for n=engine.postorder 
  for c=engine.preorder_children{n}
    % seppot{n,c} = marginalize_pot(clpot{n}, engine.separator{n,c}, engine.maximize);
    engine.marg_cl_onto_sep_ndx_id(n,c) = add_ndx(engine.cliques{n}, engine.separator{n,c}, ns, ndx_type);
    % clpot{c} = multiply_by_pot(clpot{c}, seppot{n,c});
    engine.mult_cl_by_sep_ndx_id(c,n) = add_ndx(engine.cliques{c},  engine.separator{n,c}, ns, ndx_type);
  end
end

% init pot / marginal_family
for n=1:N
  c = engine.clq_ass_to_node(n);
  engine.mult_node_ndx_id(n) = add_ndx(engine.cliques{c}, family(bnet.dag, n), ns, ndx_type);
  engine.marg_fam_ndx_id(n) = add_ndx(engine.cliques{c}, family(bnet.dag, n), ns, ndx_type);
  engine.marg_node_ndx_id(n) = add_ndx(engine.cliques{c}, n, ns, ndx_type);
end

  

%%%%%%%%

function engine = init_fields()

engine.jtree = [];
engine.cliques = [];
engine.separator = [];
engine.cliques_bitv = [];
engine.clique_weight = [];
engine.clpot = [];
engine.clq_ass_to_node = [];
engine.root_clq = [];
engine.preorder = [];
engine.postorder = [];
engine.preorder_children = [];
engine.postorder_parents = [];
engine.maximize = [];
engine.evidence = [];
engine.actual_node_sizes = [];
engine.eff_node_sizes = [];
engine.bnet_parents = [];
engine.mult_cl_by_sep_ndx_id = [];
engine.marg_cl_onto_sep_ndx_id = [];
engine.mult_node_ndx_id = [];
engine.marg_fam_ndx_id = [];
engine.marg_node_ndx_id = [];
engine.ndx_type = [];
