function engine = stab_cond_gauss_inf_engine(bnet, varargin)
% STAB_COND_GAUSS_INF_ENGINE Junction tree using stable CG potentials
% engine = cond_gauss_inf_engine(bnet, ...)
% 
% The following optional arguments can be specified in the form of name/value pairs:
% [default value in brackets]
%
% clusters  - a cell array of sets of nodes we want to ensure are in the same clique (in addition to families) [ {} ]
% root      - the root of the junction tree will be a clique that contains this node [N]
% stages    - stages{t} is a set of nodes we want to eliminate before stages{t+1}, ... [ {1:N} ]
%
% e.g., engine = stab_cond_gauss_inf_engine(bnet, 'maximize', 1);
%
% This class was written by Shan Huang (shan.huang@intel.com)

% set default params
N = length(bnet.dag);
clusters = {};
root = N;
stages = { 1:N };

if nargin >= 2
  args = varargin;
  nargs = length(args);
  for i=1:2:nargs
    switch args{i},
     case 'clusters', clusters = args{i+1}; 
     case 'root',     root = args{i+1}; 
     case 'stages',   stages = args{i+1}; 
     otherwise,  
      error(['invalid argument name ' args{i}]);       
    end
  end
end

engine.evidence = [];

[engine.jtree, dummy, engine.cliques, B, w] = dag_to_jtree(bnet, bnet.observed, stages, clusters);
engine.cliques_bitv = B;
engine.clique_weight = w;

% find a clique containing the root node
root_clq = 0;
cl = engine.cliques;
for i=1:length(cl)
  if mysubset(root, cl{i})
    root_clq = i;
    break;
  end
end
assert(root_clq > 0);

% Make the jtree rooted, so there is a fixed message passing order.
[engine.jtree, engine.preorder, engine.postorder] = mk_rooted_tree(engine.jtree, root_clq);
engine.root = root_clq;

% A node can be a member of many cliques, but is assigned to exactly one, to avoid
% double-counting its CPD. We assign node i to clique c if c is the "lightest" clique that
% contains i's family, so it can accomodate its CPD.

engine.clq_ass_to_node = zeros(1, N);
num_cliques = length(engine.cliques);
for i=1:N
  clqs_containing_family = find(all(B(:,family(bnet.dag, i)), 2)); % all selected columns must be 1
  c = clqs_containing_family(argmin(w(clqs_containing_family)));  
  engine.clq_ass_to_node(i) = c; 
end

% Compute the separators between connected cliques.

[is,js] = find(engine.jtree > 0);
engine.separator = cell(num_cliques, num_cliques);
for k=1:length(is)
  i = is(k); j = js(k);
  engine.separator{i,j} = find(B(i,:) & B(j,:)); % intersect(cliques{i}, cliques{j});
end


C = length(engine.cliques);
engine.clpot = cell(1,C);
engine.seppot = cell(C,C);
ns = bnet.node_sizes(:);
N = length(bnet.dag);

pot_type = 'scg';
check_for_cd_arcs([], bnet.cnodes, bnet.dag);

% Evaluate CPDs with evidence, and convert to potentials  
pot = cell(1, N);
C = length(engine.cliques);
inited = zeros(1, C);
clpot = cell(1, C);
evidence = cell(1, N);
for n=1:N
  fam = family(bnet.dag, n);
  e = bnet.equiv_class(n);
  %pot{n} = CPD_to_scgpot(bnet.CPD{e}, fam, ns, bnet.cnodes, evidence);
  pot{n} = convert_to_pot(bnet.CPD{e}, pot_type, fam(:), evidence);
  cindex = engine.clq_ass_to_node(n);
  if inited(cindex)
      clpot{cindex} = direct_combine_pots(pot{n}, clpot{cindex});
  else
      clpot{cindex} = pot{n};
      inited(cindex) = 1;
  end
end

for i=1:C
    if inited(i) == 0
        clpot{i} = scgpot([], [], [], []);
    end
end

seppot = cell(C, C);
% separators are is not need to initialize

% collect to root (node to parents)
% Unlike the HUGIN architecture the complements are stored in the cliques during COLLECT 
% and the separators are not playing a specific role during this process


engine.clpot = clpot;
engine.seppot = seppot;

engine = class(engine, 'stab_cond_gauss_inf_engine', inf_engine(bnet));
