function CPDpot = convert_dbn_CPDs_to_tables(bnet, evidence)
% CONVERT_DBN_CPDS_TO_TABLES Convert CPDs of (possibly instantiated) DBN nodes to tables
% CPDpot = convert_dbn_CPDs_to_tables(bnet, evidence)
%
% CPDpot{n,t} is a table containing P(n,t|pa(n,t), ev)
% All hidden nodes are assumed to be discrete
% We assume the observed nodes are the same in every slice
%
% Evaluating the conditional likelihood of the evidence can be very slow,
% so we take pains to vectorize where possible

[ss T] = size(evidence);
%obs_bitv = ~isemptycell(evidence(:));
obs_bitv = zeros(1, 1:2*ss);
obs_bitv(bnet.observed) = 1;
obs_bitv(bnet.observed+ss) = 1;

ns = bnet.node_sizes(:);
CPDpot = cell(ss,T); 

for n=1:ss
  
% slice 1
t = 1;
ps = parents(bnet.dag, n);
e = bnet.equiv_class(n, 1);
if ~any(obs_bitv(ps))
  CPDpot{n,t} = convert_CPD_to_table_hidden_ps(bnet.CPD{e}, evidence{n,t});
else
  CPDpot{n,t} = convert_to_table(bnet.CPD{e}, [ps n], evidence(:,1));
end

verbose = 1;

% slices 2..T
self = n+ss;
ps = parents(bnet.dag, self);
e = bnet.equiv_class(n, 2);
if ~obs_bitv(n) % self is hidden (hence discrete)
  if ~any(obs_bitv(ps)) % all parents are hidden (hence hidden)
    if verbose, fprintf('node %d is hidden, all ps are hidden\n', n); end
    if myismember(n, bnet.dnodes) 
      %CPT = CPD_to_CPT(bnet.CPD{e});
      %CPT = reshape(CPT, [prod(ns(ps)) ns(self)]);
      CPT = convert_CPD_to_table_hidden_ps(bnet.CPD{e}, []);
      CPDpot(n,2:T) = num2cell(repmat(CPT, [1 1 T-1]), [1 2]);
    else
      error(['hidden cts node disallowed'])
    end
  else % some parents are observed - slow
    % We assume observed parents are discrete - not always true
    if verbose, fprintf('node %d is hidden, some ps are obs\n', n); end
    domain = [ps self];
    odom = find(obs_bitv(domain));
    CPT = CPD_to_CPT(bnet.CPD{e});
    map = find_equiv_posns(odom, domain);
    index = cell(1, D);
    for t=2:T
      %CPDpot{n,t} = convert_to_table(bnet.CPD{e}, [ps self], evidence(:,t-1:t));
      ev = evidence(:, t-1:t);
      vals = cat(1, ev{odom});
      index = mk_multi_index(length(domain), map, vals);
      T = CPT(index{:});
      CPDpot(n,t) = T(:);
    end
  end
else % self is observed
  if ~any(obs_bitv(ps)) % all parents are hidden
    parents_in_same_slice = all(ps > ss);
    if parents_in_same_slice
      if verbose, fprintf('node %d is obs, all ps are hidden, 1 slice\n', n); end
      ps1 = ps - ss;
      if myismember(n, bnet.dnodes) 
	CPT = CPD_to_CPT(bnet.CPD{e});
	CPT = reshape(CPT, [prod(ns(ps)) ns(self)]); % what if no parents?
	obslik = mk_dhmm_obs_lik(cell2num(evidence(n,2:T)), CPT);
	CPDpot(n,2:T) = num2cell(obslik, 1);
      else
	S = struct(bnet.CPD{e}); 
	obslik = mk_ghmm_obs_lik(cell2num(evidence(n,2:T)), S.mean, S.cov);
	CPDpot(n,2:T) = num2cell(obslik, 1);
      end
    else % parents span 2 slices - slow
      if verbose, fprintf('node %d is obs, all ps are hidden , 2 slice\n', n); end
      for t=2:T
	CPDpot{n,t} = convert_to_table(bnet.CPD{e}, [ps self], evidence(:,t-1:t));
      end
    end
  else % some parents are observed - slow
    if verbose, fprintf('node %d is obs, some ps are obs', n); end
    for t=2:T
      CPDpot{n,t} = convert_to_table(bnet.CPD{e}, [ps self], evidence(:,t-1:t));
    end
  end
end

end

if 0
CPDpot2 = convert_dbn_CPDs_to_tables_slow(bnet, evidence);
for t=1:T
  for n=1:ss
    if ~approxeq(CPDpot{n,t}, CPDpot2{n,t})
      fprintf('CPDpot n=%d, t=%d\n',n,t);
      keyboard
    end
  end
end
end
